#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "chess.h"
#include "data.h"
#include "epdglue.h"

#define SINGULAR_MARGIN 50

/* modified 02/02/98 */
/*
********************************************************************************
*                                                                              *
*   Singular() is a routine used by Search() to determine if a move appears to *
*   be "singular".  IE is there one move that appears to be better than all of *
*   the other moves by a significant margin (SINGULAR_MARGIN).  if so, this    *
*   move is passed back to Search() which will extend this one move by the     *
*   normal extension amount (3/4 ply).                                         *
*                                                                              *
********************************************************************************
*/
int Singular(int alpha, int wtm, int depth, int ply)
{
  register int singular=0;
  int singular_move=0, tried=0;
  register int value;
  register int save_hash_move;
/*
 ----------------------------------------------------------
|                                                          |
|   initialize.                                            |
|                                                          |
 ----------------------------------------------------------
*/
  in_check[ply+1]=0;
  extended_reason[ply+1]=no_extension;
  last[ply]=last[ply-1];
  killer_count1[ply+1]=0;
  killer_count2[ply+1]=0;
  next_status[ply].phase=FIRST_PHASE;
  save_hash_move=hash_move[ply];
printf("starting singular test, ply=%d\n",ply);
/*
 ----------------------------------------------------------
|                                                          |
|   now iterate through the move list and search the       |
|   resulting positions.  note that Search() culls any     |
|   move that is not legal by using Check().  the special  |
|   case is that we must find one legal move to search to  |
|   confirm that it's not a mate or draw.                  |
|                                                          |
 ----------------------------------------------------------
*/
  while ((current_phase[ply]=(in_check[ply]) ? NextEvasion(ply,wtm) : 
                                               NextMove(ply,wtm))) {
    if (++tried>=3 && singular==0) break;
#if !defined(FAST)
    if (ply <= trace_level)
      SearchTrace(ply,depth,wtm,alpha-SINGULAR_MARGIN-1,
                  alpha-SINGULAR_MARGIN,"Singular",current_phase[ply]);
#endif
/*
 ----------------------------------------------------------
|                                                          |
|   if there's only one legal move, the move is singular   |
|   "by definition".                                       |
|                                                          |
 ----------------------------------------------------------
*/
    if (last[ply]-last[ply-1] == 1) {
      singular=2;
      break;
    }
/*
 ----------------------------------------------------------
|                                                          |
|   now make the move and search the resulting position.   |
|   if we are in check, the current move must be legal     |
|   since NextEvasion ensures this, otherwise we have to   |
|   make sure the side-on-move is not in check after the   |
|   move to weed out illegal moves and save time.          |
|                                                          |
 ----------------------------------------------------------
*/
    MakeMove(ply,current_move[ply],wtm);
    if (in_check[ply] || !Check(wtm)) {
/*
 ----------------------------------------------------------
|                                                          |
|   if the move to be made checks the opponent, then we    |
|   need to remember that he's in check.                   |
|                                                          |
 ----------------------------------------------------------
*/
      if (Check(ChangeSide(wtm))) in_check[ply+1]=1;
      else in_check[ply+1]=0;
      value=-ABSearch(-alpha+SINGULAR_MARGIN+1,-alpha+SINGULAR_MARGIN,
                      ChangeSide(wtm), depth+240,ply+1,1);
      if (abort_search) {
        UnMakeMove(ply,current_move[ply],wtm);
        return(0);
      }
      if (value > alpha-SINGULAR_MARGIN-1) {
        singular++;
        singular_move=current_move[ply];
      }
    }
    UnMakeMove(ply,current_move[ply],wtm);
    if (singular > 1) break;
  }
/*
 ----------------------------------------------------------
|                                                          |
|   all moves have been searched.                          |
|                                                          |
 ----------------------------------------------------------
*/
  hash_move[ply]=save_hash_move;
  if (singular != 1) return(0);
#if !defined(FAST)
    if (ply <= trace_level)
      printf("ply=%d move %s appears to be 'singular'\n",
             ply,OutputMove(&singular_move,ply,wtm));
#endif
  return(singular_move);
}
